/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <alloca.h>

#include "config.h"
#include "huffman.h"

/* Gives the number of code words of a given bit length, the first
   code word, and the index of that code word. Code words are selected
   and ordered according to RFC 1951 (lexicographic order, shorter
   codes precede longer, codes of the same length are consecutive). */
struct code_info {
  unsigned short count;
  unsigned short index;
  unsigned short code;
};

/* Locates the largest code word, of length at most bits, that matches
   a given prefix. */
static void
huffman_largest_code (const struct code_info *info,
		      unsigned bits,
		      unsigned prefix_length, unsigned prefix,
		      unsigned *code_length, unsigned *code)
{
  assert(prefix_length > 0);

  /* Find the largest code word matching the prefix. */
  for (; bits >= prefix_length; bits--)
    if (info[bits].count)
      {
	/* The largest word of length bits that matches the prefix. */
	unsigned prefix_max = ((prefix+1) << (bits - prefix_length)) - 1;

	if (info[bits].code <= prefix_max)
	  {
	    unsigned index;

	    /* Index relative to the first code of this length. */
	    index = prefix_max - info[bits].code;
	    if (index >= info[bits].count)
	      index = info[bits].count - 1;

	    *code_length = bits;
	    *code = info[bits].code + index;

	    return;
	  }
      }
  abort();
}

/* Construct a huffman tree representing the codes starting with a
   given prefix. BITS and LAST_CODE are the last code word that
   matches the prefix. Returns the number of used nodes. */
static unsigned
huffman_build_tree(unsigned short *tree,
		   const struct code_info *info,
		   /* VALUE[k] gives the value for the k'th code word. */
		   const unsigned short *value,
		   unsigned prefix_length,
		   unsigned bits, unsigned largest_code)
{
  unsigned zeros;

  assert (prefix_length <= bits);

  /* We have the largest code word matching the prefix. If the next
     bit of this code word is zero, that means that there's no choice,
     the zero is mandatory. */
  zeros = 0;

  while (prefix_length < bits
	 && !((largest_code >> (bits - prefix_length - 1)) & 1))
    {
      zeros++;
      prefix_length++;
    }

  if (bits == prefix_length)
    {
      /* A leaf node */
      *tree = (HUFFMAN_LEAF
	       | (zeros << HUFFMAN_ZEROS_SHIFT)
	       | value[info[bits].index + largest_code - info[bits].code]);

      return 1;
    }
  else
    {
      unsigned prefix;
      unsigned bits_left;
      unsigned largest_left;
      unsigned left;
      unsigned right;

      prefix = largest_code >> (bits - prefix_length);
      
      huffman_largest_code (info, bits, prefix_length + 1, prefix << 1,
			    &bits_left, &largest_left);
      
      left = huffman_build_tree(tree + 1,
				info, value,
				prefix_length + 1,
				bits_left, largest_left);
      
      right = huffman_build_tree(tree + 1 + left,
				 info, value,
				 prefix_length + 1,
				 bits, largest_code);

      *tree = (zeros << HUFFMAN_ZEROS_SHIFT) | left;
      return 1 + left + right;
    }
}

#define MAX_LENGTH 15

int
huffman_tree(unsigned short *tree,
	     unsigned n, const unsigned char *length)
{
  struct code_info info[MAX_LENGTH + 1];
  unsigned short index[MAX_LENGTH + 1];
  unsigned bits;
  unsigned k;
  
  /* Gives the value for the j'th code word, with code words ordered
     lexicographically. */
  unsigned short *value;
  unsigned size;

  assert (n > 0);
  memset(info, 0, sizeof(info));

  for (k = 0; k < n; k++)
    {
      if (length[k] > MAX_LENGTH)
	return 0;
      info[length[k]].count++;
    }

  for (bits = MAX_LENGTH; !info[bits].count; bits--)
    ;

  if (!bits)
    /* No code words? All specified to have length zero. */
    return 0;

  /* NOTE: The algorithm in RFC 1951 seem to have an off-by-one error.
     We should not allocate any codes of length zero. */
  info[1].code = 0;
  info[1].index = 0;

  for (k = 2; k <= bits; k++)
    {
      info[k].code = (info[k-1].code + info[k-1].count) << 1;
      info[k].index = info[k-1].index + info[k-1].count;

      /* Make sure code fits in given size. */
      if (info[k].code + info[k].count > (1U << k))
	return 0;
    }
  
  value = alloca((n - info[0].count)  * sizeof(*value));
  memset(index, 0, sizeof(index));

  for (k = 0; k < n; k++)
    if (length[k] > 0)
      value[info[length[k]].index + index[length[k]]++] = k;
    
  size = huffman_build_tree(tree, info, value, 0,
			    bits, info[bits].code + info[bits].count - 1);
  assert(size == HUFFMAN_SIZE(n - info[0].count));

  /* Sanity check */
  for (k = 0; k < size; k++)
    assert(HUFFMAN_IS_LEAF(tree[k])
	   || (k + 1 + HUFFMAN_GET_VALUE(tree[k]) < size));

  return 1;
}

void
huffman_init (struct huffman_ctx *ctx,
	      const unsigned short *root)
{
  ctx->root = root;
  ctx->p = root;
  ctx->zeros = HUFFMAN_GET_ZEROS(*root);

  assert(!HUFFMAN_IS_LEAF(*root) || ctx->zeros);
}

/* This function is crucial for performance, and a target for
   micro-optimization. Uses a loop around the common case zeros =
   0. */
int
huffman_decode_bits (struct huffman_ctx *ctx,
		     struct bit_buffer *input,
		     unsigned *value)
{
  const unsigned short *p = ctx->p;
  unsigned bits = input->bits;
  unsigned n = input->n;
  unsigned node;
  int res;
  
  if (UNLIKELY(ctx->zeros))
    {
    handle_zeros:
      assert(n > 0);
      
      if (n < ctx->zeros)
	{
	  if (bits & ((1 << n) - 1))
	    return -1;

	  ctx->zeros -= n;
	  ctx->p = p;
	  input->n = 0;
	  input->bits = 0;
	  return 0;
	}
      if (bits & ((1 << ctx->zeros) - 1))
	return -1;

      n -= ctx->zeros;
      ctx->zeros = 0;
      bits >>= ctx->zeros;
      
      node = *p;
      if (HUFFMAN_IS_LEAF(node))
	goto leaf;

      else if (n == 0)
	{
	  res = 0;
	  goto done;
	}
    }
  
  /* Common, time critical, case. */
  assert(n > 0);

  /* Alternative test condition, with a single branch:

     ((((node = *p) | (n - 1)) & (HUFFMAN_LEAF | HUFFMAN_ZEROS_MASK)) == 0)
  */
  while (((node = *p) & (HUFFMAN_LEAF | HUFFMAN_ZEROS_MASK)) == 0 && n > 0)
    {
      p++;

      p += node & - (bits & 1);

      bits >>= 1;
      n--;
    }

  if (UNLIKELY(node & HUFFMAN_ZEROS_MASK))
    {
      ctx->zeros = HUFFMAN_GET_ZEROS(node);
      if (n > 0)
	goto handle_zeros;
      else
	res = 0;
    }
  else if (HUFFMAN_IS_LEAF(node))
    {
    leaf:
      *value = HUFFMAN_GET_VALUE(node);
      res = 1;
	  
      p = ctx->root;
      ctx->zeros = HUFFMAN_GET_ZEROS(*p);
    }
  else
    res = 0;

 done:
  ctx->p = p;
  input->n = n;
  input->bits = bits;

  return res;
}
